@@ -1,6 +1,6 @@
 #!/usr/bin/env python
 #
-# Copyright (c) 2024, Ryan Galloway (ryan@rsgalloway.com)
+# Copyright (c) 2024-2025, Ryan Galloway (ryan@rsgalloway.com)
 #
 # Redistribution and use in source and binary forms, with or without
 # modification, are permitted provided that the following conditions are met:
@@ -34,6 +34,7 @@ Command line interface for envstack: stacked environment variable management.
 """
 
 import argparse
+import re
 import sys
 import traceback
 
@@ -54,6 +55,74 @@ from envstack.wrapper import run_command
 setup_stream_handler()
 
 
+class StoreOnce(argparse.Action):
+    """Custom argparse action to ensure an option is only set once."""
+
+    def __call__(self, parser, namespace, values, option_string=None):
+        # if we've already seen this option once, bail
+        if getattr(namespace, f"__seen_{self.dest}", False):
+            aliases = "/".join(self.option_strings)
+            parser.error(f"{aliases} specified more than once.")
+        setattr(namespace, f"__seen_{self.dest}", True)
+        setattr(namespace, self.dest, values)
+
+
+def _parse_env_lines(lines):
+    """Parse lines of environment variables from an iterable.
+
+    :param lines: An iterable of lines, such as a file or stdin.
+    :return: A dictionary of environment variables.
+    """
+    ENV_LINE_RE = re.compile(
+        r"""
+        ^\s*
+        (?:export\s+)?
+        (?P<key>(?:<<|[A-Za-z_][A-Za-z0-9_]*))
+        \s*
+        (?P<op>[:=])
+        \s*
+        (?P<val>.*)
+        \s*$
+        """,
+        re.VERBOSE,
+    )
+
+    def parse_line(line: str):
+        m = ENV_LINE_RE.match(line)
+        if not m:
+            return None
+        key = m.group("key")
+        val = m.group("val")
+        # strip matching quotes only if present
+        if len(val) >= 2 and val[0] == val[-1] and val[0] in ("'", '"'):
+            val = val[1:-1]
+        return key, val
+
+    data = {}
+    for raw in lines:
+        k, v = parse_line(raw.strip())
+        data[k] = v
+    return data
+
+
+def _parse_keyvals(items: dict):
+    """Parse a list of key=value pairs.
+
+    :param items: A list of strings in the form "key=value" or "key:value".
+    :return: A dictionary mapping keys to values.
+    """
+    out = {}
+    for kv in items:
+        if ":" in kv:
+            k, v = kv.split(":", 1)
+        elif "=" in kv:
+            k, v = kv.split("=", 1)
+        else:
+            k, v = kv, ""
+        out[k] = v
+    return out
+
+
 def parse_args():
     """Command line argument parser.
 
@@ -135,8 +204,15 @@ def parse_args():
         "-s",
         "--set",
         nargs="*",
-        metavar="VAR:VALUE",
-        help="set a key:value pair in the environment",
+        action=StoreOnce,
+        metavar="KEY=VALUE",
+        help="overlay KEY=VALUE pairs to envstack environments",
+    )
+    parser.add_argument(
+        "-b",
+        "--bare",
+        action="store_true",
+        help="create a bare environment",
     )
     parser.add_argument(
         "--scope",
@@ -147,6 +223,7 @@ def parse_args():
         "-r",
         "--resolve",
         nargs="*",
+        action=StoreOnce,
         metavar="VAR",
         help="resolve environment variables",
     )
@@ -154,6 +231,7 @@ def parse_args():
         "-t",
         "--trace",
         nargs="*",
+        action=StoreOnce,
         metavar="VAR",
         help="trace where environment variables are being set",
     )
@@ -162,6 +240,12 @@ def parse_args():
         action="store_true",
         help="list the env stack file sources",
     )
+    export_group.add_argument(
+        "-q",
+        "--quiet",
+        action="store_true",
+        help="print values only",
+    )
 
     args = parser.parse_args(args_before_dash)
 
@@ -203,18 +287,61 @@ def main():
                 for key, value in data.items():
                     print(f"{key}={value}")
 
-        elif args.set:
-            data = dict(kv.split(":", 1) for kv in args.set)
+        elif args.set is not None:
+            force_stdin = args.set == [] or args.set == ["-"]
+            using_pipe = args.set == [] and not sys.stdin.isatty()
 
-            if args.encrypt:
-                data = encrypt_environ(data, encrypt=(not args.out))
-            if args.export:
-                print(export_env_to_shell(data))
-            elif args.out:
-                Env(data).write(args.out)
+            # load the environment if not in bare mode
+            if args.bare:
+                env = Env()
             else:
-                for key, value in data.items():
-                    print(f"{key}={value}")
+                env = load_environ(args.namespace, platform=args.platform)
+
+            # interactive mode
+            if force_stdin and sys.stdin.isatty():
+                print(
+                    "Enter ENV:VAR or ENV=VAR pairs. Press Ctrl+D or Ctrl+C to finish:",
+                    file=sys.stderr,
+                )
+                lines = []
+                try:
+                    while True:
+                        lines.append(input() + "\n")
+                except (EOFError, KeyboardInterrupt):
+                    pass
+                data = _parse_env_lines(lines)
+                if not data:
+                    return 0
+            # pipe stdin (or '-' with non-tty stdin)
+            elif force_stdin or using_pipe:
+                data = _parse_env_lines(sys.stdin)
+                if not data:
+                    raise ValueError("no KEY=VALUE pairs found on stdin")
+            # explicit args path
+            else:
+                data = _parse_keyvals(args.set)
+
+            # encrypt the new data only
+            if args.encrypt:
+                data = encrypt_environ(data)
+
+            # update the environment with the new data
+            env.update(data)
+
+            if args.export:
+                print(export_env_to_shell(env))
+            elif args.out:
+                Env(env).write(args.out)
+            else:
+                for key, val in env.items():
+                    if args.quiet:
+                        if len(env) > 1:
+                            print("error: --quiet requires exactly one KEY")
+                            return 2
+                        else:
+                            print(val)
+                    else:
+                        print(f"{key}={val}")
 
         elif args.out:
             bake_environ(
@@ -231,14 +358,30 @@ def main():
             keys = args.resolve or resolved.keys()
             for key in sorted(str(k) for k in keys):
                 val = resolved.get(key)
-                print(f"{key}={val}")
+                if key in resolved:
+                    if args.quiet:
+                        if len(keys) > 1:
+                            print("error: --quiet requires exactly one KEY")
+                            return 2
+                        else:
+                            print(val)
+                    else:
+                        print(f"{key}={val}")
 
         elif args.trace is not None:
             if len(args.trace) == 0:
                 args.trace = load_environ(args.namespace).keys()
             for trace in args.trace:
                 path = trace_var(*args.namespace, var=trace)
-                print("{0}: {1}".format(trace, path))
+                if path:
+                    if args.quiet:
+                        if len(args.trace) > 1:
+                            print("error: --quiet requires exactly one KEY")
+                            return 2
+                        else:
+                            print(path)
+                    else:
+                        print("{0}={1}".format(trace, path))
 
         elif args.sources:
             env = load_environ(args.namespace, platform=args.platform)
