@@ -1,6 +1,6 @@
 #!/usr/bin/env python
 #
-# Copyright (c) 2024, Ryan Galloway (ryan@rsgalloway.com)
+# Copyright (c) 2024-2025, Ryan Galloway (ryan@rsgalloway.com)
 #
 # Redistribution and use in source and binary forms, with or without
 # modification, are permitted provided that the following conditions are met:
@@ -41,15 +41,17 @@ from pathlib import Path
 import yaml  # noqa
 
 from envstack import config, logger, path, util
-from envstack.node import custom_node_types, get_keys_from_env, BaseNode, EncryptedNode
 from envstack.exceptions import *  # noqa
+from envstack.node import (
+    BaseNode,
+    EncryptedNode,
+    custom_node_types,
+    get_keys_from_env,
+)
 
 # value delimiter pattern (splits values by os.pathsep)
 delimiter_pattern = re.compile("(?![^{]*})[;:]+")
 
-# stores cached file data in memory
-load_file_cache = {}
-
 # stores environment when calling envstack.save()
 saved_environ = None
 
@@ -164,7 +166,8 @@ class EnvVar(string.Template, str):
         self.template = v
 
     def extend(self, iterable):
-        """If value is a list, extend list by appending elements from the iterable.
+        """If value is a list, extend list by appending elements from the
+        iterable.
 
         :param iterable: an iterable object
         :returns: extended value
@@ -214,12 +217,12 @@ class EnvVar(string.Template, str):
         return self.template.keys()
 
     def parts(self):
-        """Returns a list of delimited parts, e.g. if a value is delimited by a colon
-        (or semicolon on windows), e.g. ::
+        """Returns a list of delimited parts, e.g. if a value is delimited by a
+        colon (or semicolon on windows), for example:
 
-        >>> v = EnvVar('$FOO:${BAR}/bin')
-        >>> v.parts()
-        ['$FOO', '${BAR}/bin']
+            >>> v = EnvVar('$FOO:${BAR}/bin')
+            >>> v.parts()
+            ['$FOO', '${BAR}/bin']
         """
         if self.template:
             if type(self.template) in (
@@ -341,7 +344,10 @@ class Env(dict):
 
         # write the baked environment to the file
         if filename:
-            baked.write()
+            try:
+                baked.write()
+            except Exception as err:
+                raise WriteError(f"Failed to write {filename}", err)
 
         # create the baked environment from the baked source
         baked_env = Env()
@@ -377,12 +383,7 @@ class Env(dict):
             return source
 
 
-def clear_file_cache():
-    """Clears global file cache."""
-    global load_file_cache
-    load_file_cache = {}
-
-
+@util.cache
 def get_sources(
     *names,
     scope: str = None,
@@ -399,16 +400,12 @@ def get_sources(
     :raises TemplateNotFound: if a file is not found in ENVPATH or scope.
     :returns: list of Source objects for the given stack names.
     """
-    # TODO: smarter file caching (issue #26)
-    # clear_file_cache()
-
-    # set default scope to the current working directory
-    scope = Path(scope or os.getcwd()).resolve()
-
     loaded_files = []
     sources = []
     loading_stack = set()
 
+    scope = Path(scope or os.getcwd()).resolve()
+
     def _walk_to_scope(current_path):
         """Generate directories from the current path up to the scope."""
         paths = []
@@ -432,8 +429,15 @@ def get_sources(
         found_files = []
         for directory in search_paths:
             potential_file = Path(directory) / file_basename
-            if potential_file.exists() and potential_file not in found_files:
-                found_files.append(potential_file)
+            try:
+                if potential_file.exists() and potential_file not in found_files:
+                    found_files.append(potential_file)
+            except PermissionError:
+                logger.log.warning(f"Permission denied: {potential_file}")
+                continue
+            except OSError:
+                logger.log.warning(f"Error accessing {potential_file}")
+                continue
         if not found_files and not ignore_missing:
             raise TemplateNotFound(  # noqa
                 f"{file_basename} not found in ENVPATH or scope."
@@ -722,6 +726,9 @@ def init(*name, ignore_missing: bool = config.IGNORE_MISSING):
 
     # load the stack and update the environment
     env = resolve_environ(load_environ(name, ignore_missing=ignore_missing))
+
+    # os.environ expects strings, so encode the values
+    # TypeError: str expected, not int
     os.environ.update(util.encode(env))
 
     # update sys.path from resolved PYTHONPATH
@@ -787,6 +794,7 @@ def encrypt_environ(
             node = node_class(v)
             if encrypt:
                 node.value = node.encryptor(env=resolved_env).encrypt(str(v))
+                node.original_value = node.value
             encrypted_env[k] = node
         else:
             encrypted_env[k] = v
@@ -794,9 +802,11 @@ def encrypt_environ(
     return encrypted_env
 
 
-def resolve_environ(env: dict):
-    """Resolves all variables in a given unresolved environment, returning a
-    new environment dict.
+def resolve_environ(env: Env):
+    """Resolves all variables in a given environment, returning a new
+    environment with resolved values.
+
+        >>> env = resolve_environ(load_environ(name))
 
     :param env: unresolved environment.
     :returns: resolved environment.
@@ -804,26 +814,43 @@ def resolve_environ(env: dict):
     # stores the resolved environment
     resolved = Env()
 
+    if type(env) is not Env:
+        env = Env(env)
+
     # copy env to avoid modifying the original
     env_copy = env.copy()
 
+    # evaluate one source environment at a time
+    seen_source_paths = []
+    included = Env()
+    sources = env.sources[:-1]
+    sources.reverse()
+    for source in sources:
+        if source.path in seen_source_paths:
+            continue
+        seen_source_paths.append(source.path)
+        source_environ = resolve_environ(Env(source.load()))
+        for key, value in source_environ.items():
+            included[key] = util.evaluate_modifiers(value, environ=source_environ)
+
     # make a copy that contains the encryption keys
     env_keys = env.copy()
     env_keys.update(get_keys_from_env(os.environ))
 
-    # decrypt custom node types
-    for key, value in env_copy.items():
+    # decrypt custom nodes
+    for key, value in list(env_copy.items()):
         if type(value) in custom_node_types:
             env_copy[key] = value.resolve(env=env_keys)
 
-    # resolve variables after decrypting custom node types
-    for key, value in env_copy.items():
-        resolved[key] = util.evaluate_modifiers(value, environ=env_copy)
+    # resolve environment variables after decrypting custom nodes
+    for key, value in list(env_copy.items()):
+        value = util.evaluate_modifiers(value, environ=env_copy, parent=included)
+        resolved[key] = util.safe_eval(value)
 
     return resolved
 
 
-# TODO: make 'name' arg a list (*names) in next minor release
+# TODO: make 'name' arg a list (*names)
 def load_environ(
     name: str = config.DEFAULT_NAMESPACE,
     platform: str = config.PLATFORM,
@@ -831,8 +858,8 @@ def load_environ(
     ignore_missing: bool = config.IGNORE_MISSING,
     encrypt: bool = False,
 ):
-    """Loads env stack data for a given name. Adds "STACK" key to environment,
-    and sets the value to `name`.
+    """Loads raw environment for a given stack namespace.
+    Adds "STACK" key to environment, and sets the value to `name`.
 
     To load an environment for a given namespace, where the scope is the current
     working directory (cwd):
@@ -843,6 +870,10 @@ def load_environ(
 
         >>> env = load_environ(name, scope="/path/to/scope")
 
+    To load resolved environment, use `resolve_environ()`:
+
+        >>> env = resolve_environ(load_environ(name))
+
     :param name: list of stack names to load (basename of env files).
     :param platform: name of platform (linux, darwin, windows).
     :param scope: environment scope (default: cwd).
@@ -876,20 +907,20 @@ def load_environ(
             ignore_missing=ignore_missing,
         )
         for source in sources:
-            # dedupe sources (may override previous sources)
+            # don't load the same source more than once
             if source.path in seen_paths:
                 continue
             env.load_source(source, platform=platform)
             seen_paths.append(source.path)
 
+            # add the stack name to the environment
+            if not env.get("STACK"):
+                env["STACK"] = util.get_stack_name(name)
+
         # resolve ${ENVPATH} (don't let it be None)
         # TODO: use expandvars() instead of resolve_environ()
         envpath = resolve_environ(env).get("ENVPATH", envpath) or envpath
 
-    # add the stack name to the environment
-    if not env.get("STACK"):
-        env["STACK"] = util.get_stack_name(name)
-
     # encrypt the values in the environment last
     if encrypt:
         return encrypt_environ(env)
@@ -897,27 +928,22 @@ def load_environ(
     return env
 
 
+@util.cache
 def load_file(path: str):
-    """Reads a given .env file and returns data as dict.
+    """Reads a given .env file and returns data as dict. Data is memoized in
+    memory for faster access.
 
-    :param path: path to envstack env file.
+    :param path: path to .env file.
     :returns: loaded yaml data as dict.
     """
-    global load_file_cache
-
-    if path in load_file_cache:
-        return load_file_cache[path]
 
     if not os.path.exists(path):
         return {}
 
-    else:
-        data = util.validate_yaml(path)
-        load_file_cache[path] = data
-
-    return data
+    return util.validate_yaml(path)
 
 
+@util.cache
 def trace_var(*name, var: str = None, scope: str = None):
     """Traces where a var is getting set for a given name.
 
