@@ -1,6 +1,6 @@
 #!/usr/bin/env python
 #
-# Copyright (c) 2024, Ryan Galloway (ryan@rsgalloway.com)
+# Copyright (c) 2024-2025, Ryan Galloway (ryan@rsgalloway.com)
 #
 # Redistribution and use in source and binary forms, with or without
 # modification, are permitted provided that the following conditions are met:
@@ -34,9 +34,11 @@ Contains unit tests for running commands.
 """
 
 import os
+import platform
 import shutil
 import subprocess
 import sys
+import tempfile
 import unittest
 
 import envstack
@@ -45,6 +47,22 @@ from envstack.encrypt import AESGCMEncryptor, FernetEncryptor
 from test_env import create_test_root, update_env_file
 
 
+def make_command(envstack_bin: str, filename: str, *args: str):
+    """
+    Build a cross-platform shell command that runs envstack (with args)
+    and then prints the output file.
+    """
+
+    envstack_cmd = f'{envstack_bin} {" ".join(args)}'
+    if filename:
+        envstack_cmd += f' -o "{filename}"'
+
+    if os.name == "nt" or platform.system().lower().startswith("win"):
+        return f'{envstack_cmd} & type "{filename}"'
+    else:
+        return f'{envstack_cmd}; cat "{filename}"'
+
+
 class TestUnresolved(unittest.TestCase):
     """Tests unresolved environment variables from the cli."""
 
@@ -59,7 +77,6 @@ class TestUnresolved(unittest.TestCase):
             "darwin": "/Volumes/pipe",
         }.get(sys.platform)
         os.environ["ENVPATH"] = envpath
-        os.environ["INTERACTIVE"] = "0"
 
     def test_default(self):
         expected_output = (
@@ -169,7 +186,6 @@ class TestEncrypt(unittest.TestCase):
             "darwin": "/Volumes/pipe",
         }.get(sys.platform)
         os.environ["ENVPATH"] = envpath
-        os.environ["INTERACTIVE"] = "0"
         os.environ["ROOT"] = "/var/tmp/pipe"  # ROOT cannot be overridden
         # remove so we use base64 encoding by default
         if AESGCMEncryptor.KEY_VAR_NAME in os.environ:
@@ -271,7 +287,6 @@ class TestResolved(unittest.TestCase):
             "darwin": "/Volumes/pipe",
         }.get(sys.platform)
         os.environ["ENVPATH"] = envpath
-        os.environ["INTERACTIVE"] = "0"
         os.environ["ROOT"] = "/var/tmp/pipe"  # ROOT cannot be overridden
 
     def test_default(self):
@@ -295,7 +310,7 @@ STACK=dev
         self.assertEqual(output, expected_output)
 
     def test_distman(self):
-        expected_output = f"""DEPLOY_ROOT={self.root}/prod
+        expected_output = f"""DEPLOY_ROOT={self.root}/{os.getenv('ENV', 'prod')}
 ROOT={self.root}
 STACK=distman
 """
@@ -367,7 +382,6 @@ class TestBake(unittest.TestCase):
             "darwin": "/Volumes/pipe",
         }.get(sys.platform)
         os.environ["ENVPATH"] = envpath
-        os.environ["INTERACTIVE"] = "0"
         # remove so we use base64 encoding by default
         if AESGCMEncryptor.KEY_VAR_NAME in os.environ:
             del os.environ[AESGCMEncryptor.KEY_VAR_NAME]
@@ -380,11 +394,10 @@ class TestBake(unittest.TestCase):
 
     def test_default(self):
         """Tests baking the default stack."""
-        command = f"%s -o {self.filename}; cat {self.filename}" % self.envstack_bin
+        command = make_command(self.envstack_bin, self.filename)
         expected_output = """#!/usr/bin/env envstack
 include: []
 all: &all
-  <<: *all
   DEPLOY_ROOT: ${ROOT}/${ENV}
   ENV: prod
   ENVPATH: ${DEPLOY_ROOT}/env:${ENVPATH}
@@ -411,11 +424,10 @@ windows:
 
     def test_dev(self):
         """Tests baking the dev stack."""
-        command = f"%s dev -o {self.filename}; cat {self.filename}" % self.envstack_bin
+        command = make_command(self.envstack_bin, self.filename, "dev")
         expected_output = """#!/usr/bin/env envstack
 include: []
 all: &all
-  <<: *all
   DEPLOY_ROOT: ${ROOT}/dev
   ENV: dev
   ENVPATH: ${ROOT}/dev/env:${ROOT}/prod/env:${ENVPATH}
@@ -442,14 +454,12 @@ windows:
 
     def test_thing(self):
         """Tests baking the thing stack with depth of 1."""
-        command = (
-            f"%s thing -o {self.filename} --depth 1; cat {self.filename}"
-            % self.envstack_bin
+        command = make_command(
+            self.envstack_bin, self.filename, "thing", "--depth", "1"
         )
         expected_output = """#!/usr/bin/env envstack
 include: [default]
 all: &all
-  <<: *all
   CHAR_LIST: [a, b, c, "${HELLO}"]
   DICT: {a: 1, b: 2, c: "${INT}"}
   FLOAT: 1.0
@@ -476,13 +486,10 @@ windows:
 
     def test_default_encrypted(self):
         """Tests baking the default stack encrypted with base64."""
-        command = (
-            f"%s --encrypt -o {self.filename}; cat {self.filename}" % self.envstack_bin
-        )
+        command = make_command(self.envstack_bin, self.filename, "--encrypt")
         expected_output = """#!/usr/bin/env envstack
 include: []
 all: &all
-  <<: *all
   DEPLOY_ROOT: !encrypt JHtST09UfS8ke0VOVn0=
   ENV: !encrypt cHJvZA==
   ENVPATH: !encrypt JHtERVBMT1lfUk9PVH0vZW52OiR7RU5WUEFUSH0=
@@ -509,14 +516,12 @@ windows:
 
     def test_thing_encrypted(self):
         """Tests baking the thing stack with depth of 1 excrypted."""
-        command = (
-            f"%s thing --encrypt -o {self.filename} --depth 1; cat {self.filename}"
-            % self.envstack_bin
+        command = make_command(
+            self.envstack_bin, self.filename, "thing", "--encrypt", "--depth", "1"
         )
         expected_output = """#!/usr/bin/env envstack
 include: [default]
 all: &all
-  <<: *all
   CHAR_LIST: !encrypt WydhJywgJ2InLCAnYycsICcke0hFTExPfSdd
   DICT: !encrypt eydhJzogMSwgJ2InOiAyLCAnYyc6ICcke0lOVH0nfQ==
   FLOAT: !encrypt MS4w
@@ -543,10 +548,7 @@ windows:
 
     def test_blank(self):
         """Tests baking a blank stack."""
-        command = (
-            f"%s doesnotexist -o {self.filename}; cat {self.filename}"
-            % self.envstack_bin
-        )
+        command = make_command(self.envstack_bin, self.filename, "doesnotexist")
         expected_output = """#!/usr/bin/env envstack
 include: []
 all: &all
@@ -580,7 +582,6 @@ class TestCommands(unittest.TestCase):
             "darwin": "/Volumes/pipe",
         }.get(sys.platform)
         os.environ["ENVPATH"] = envpath
-        os.environ["INTERACTIVE"] = "0"
 
     def test_default_echo(self):
         """Tests the default stack with an echo command."""
@@ -645,7 +646,6 @@ class TestSet(unittest.TestCase):
             "darwin": "/Volumes/pipe",
         }.get(sys.platform)
         os.environ["ENVPATH"] = envpath
-        os.environ["INTERACTIVE"] = "0"
 
     def tearDown(self):
         if os.path.exists(self.filename):
@@ -653,7 +653,7 @@ class TestSet(unittest.TestCase):
 
     def test_hello_world(self):
         """Tests setting HELLO to world."""
-        command = "%s --set HELLO:world" % self.envstack_bin
+        command = "%s --set HELLO=world --bare" % self.envstack_bin
         expected_output = "HELLO=world\n"
         output = subprocess.check_output(
             command,
@@ -664,7 +664,7 @@ class TestSet(unittest.TestCase):
 
     def test_hello_world_encrypted(self):
         """Tests setting HELLO to world encrypted."""
-        command = "%s --set HELLO:world --encrypt" % self.envstack_bin
+        command = "%s --set HELLO:world --encrypt --bare" % self.envstack_bin
         expected_output = "HELLO=d29ybGQ=\n"
         output = subprocess.check_output(
             command,
@@ -675,7 +675,7 @@ class TestSet(unittest.TestCase):
 
     def test_foo_bar(self):
         """Tests setting FOO and BAR."""
-        command = r"%s -s FOO:foo BAR:\${FOO}" % self.envstack_bin
+        command = r"%s -s FOO:foo BAR:\${FOO} --bare" % self.envstack_bin
         expected_output = "FOO=foo\nBAR=${FOO}\n"
         output = subprocess.check_output(
             command,
@@ -686,7 +686,7 @@ class TestSet(unittest.TestCase):
 
     def test_foo_bar_encrypted(self):
         """Tests setting FOO and BAR encrypted."""
-        command = r"%s -s FOO:foo BAR:\${FOO} --encrypt" % self.envstack_bin
+        command = r"%s -s FOO:foo BAR:\${FOO} --encrypt --bare" % self.envstack_bin
         expected_output = "FOO=Zm9v\nBAR=JHtGT099\n"
         output = subprocess.check_output(
             command,
@@ -697,15 +697,17 @@ class TestSet(unittest.TestCase):
 
     def test_foo_bar_bake(self):
         """Tests setting FOO and BAR and bake it out to a file."""
-        command = r"%s --set FOO:foo BAR:\${FOO} -o %s; cat %s" % (
+        command = make_command(
             self.envstack_bin,
             self.filename,
-            self.filename,
+            "--set",
+            "FOO:foo",
+            "BAR:\${FOO}",
+            "--bare",
         )
         expected_output = """#!/usr/bin/env envstack
 include: []
 all: &all
-  <<: *all
   BAR: ${FOO}
   FOO: foo
 darwin:
@@ -724,15 +726,18 @@ windows:
 
     def test_foo_bar_bake_encrypted(self):
         """Tests setting FOO and BAR and bake it out to a file with encrypted values."""
-        command = r"%s --set FOO:foo BAR:\${FOO} -eo %s; cat %s" % (
+        command = make_command(
             self.envstack_bin,
             self.filename,
-            self.filename,
+            "--set",
+            "FOO:foo",
+            "BAR:\${FOO}",
+            "--encrypt",
+            "--bare",
         )
         expected_output = """#!/usr/bin/env envstack
 include: []
 all: &all
-  <<: *all
   BAR: !encrypt JHtGT099
   FOO: !encrypt Zm9v
 darwin:
@@ -759,7 +764,6 @@ class TestVarFlow(unittest.TestCase):
         )
         envpath = os.path.join(os.path.dirname(__file__), "..", "env")
         os.environ["ENVPATH"] = envpath
-        os.environ["INTERACTIVE"] = "0"
 
     def test_default_hello(self):
         command = "%s -- echo {HELLO}" % self.envstack_bin
@@ -809,7 +813,6 @@ class TestDistman(unittest.TestCase):
             "darwin": "/Volumes/pipe",
         }.get(sys.platform)
         os.environ["ENVPATH"] = envpath
-        os.environ["INTERACTIVE"] = "0"
 
     def test_default_deploy_root(self):
         os.environ["ENV"] = "invalid"  # should not be able to override ENV
@@ -856,7 +859,6 @@ class TestIssues(unittest.TestCase):
             os.path.dirname(__file__), "..", "bin", "envstack"
         )
         os.environ["ENVPATH"] = os.path.join(self.root, "prod", "env")
-        os.environ["INTERACTIVE"] = "0"
 
     def tearDown(self):
         shutil.rmtree(self.root)
@@ -922,6 +924,88 @@ class TestIssues(unittest.TestCase):
         )
         self.assertEqual(output, expected_output)
 
+    def test_issue_55(self):
+        """Tests issue 55 missing env files using test environments.
+        Tests dynamic ${ENVPATH} values that use ${STACK} in the path,
+
+        env/dev.env:
+            ENVPATH: ${ROOT}/dev/env:${ROOT}/prod/env:${ENVPATH}
+
+        env/test.env:
+            ENVPATH: ${ROOT}/${STACK}/env:${ROOT}/prod/env
+        """
+        from envstack.env import Env
+
+        # update default.env to point to test root
+        default_env_file = os.path.join(self.root, "prod", "env", "default.env")
+        update_env_file(default_env_file, "ROOT", self.root)
+
+        # create a new test env file that only exists in our test env dir
+        test_env = os.path.join(self.root, "test_issue_55", "env")
+        os.makedirs(test_env, exist_ok=True)
+        test_env_file = os.path.join(test_env, "test_issue_55.env")
+
+        data = {"FOO": "foo", "BAR": "bar", "BAZ": self.root}
+        Env(data).write(test_env_file)
+        self.assertTrue(os.path.exists(test_env_file))
+
+        # dev.env does not use STACK in ENVPATH
+        command = "%s dev test_issue_55 --sources" % self.envstack_bin
+        expected_output = f"""{self.root}/prod/env/default.env
+{self.root}/prod/env/dev.env
+"""
+        output = subprocess.check_output(
+            command, start_new_session=True, shell=True, universal_newlines=True
+        )
+        self.assertEqual(output, expected_output)
+
+        # test.env does use STACK in ENVPATH and should include our test env
+        command = "%s test test_issue_55 --sources" % self.envstack_bin
+        expected_output = f"""{self.root}/prod/env/default.env
+{self.root}/prod/env/dev.env
+{self.root}/prod/env/test.env
+{self.root}/test_issue_55/env/test_issue_55.env
+"""
+        output = subprocess.check_output(
+            command, start_new_session=True, shell=True, universal_newlines=True
+        )
+        self.assertEqual(output, expected_output)
+
+    def test_issue_66(self):
+        """Test hanging on interactive shells."""
+        # create a hostile .bashrc that will hang if loaded
+        tmp_home = tempfile.mkdtemp()
+        bashrc_path = os.path.join(tmp_home, ".bashrc")
+        with open(bashrc_path, "w") as f:
+            f.write('echo "Sourcing TEST .bashrc (PID $$)"\n')
+            f.write('read -p "Press ENTER to continue: " _\n')
+
+        # command to run envstack without inheriting our terminal
+        # use login shell to pick up bashrc if interactive
+        cmd = [
+            "bash",
+            "-lc",
+            'envstack -- "echo hello from envstack"',
+        ]
+
+        proc = subprocess.Popen(
+            cmd,
+            stdin=subprocess.DEVNULL,  # no user input possible
+            stdout=subprocess.PIPE,
+            stderr=subprocess.PIPE,
+            env={**os.environ, "HOME": tmp_home},
+        )
+
+        err = None
+        try:
+            _, err = proc.communicate(timeout=5)
+        except subprocess.TimeoutExpired:
+            proc.kill()
+            self.fail("envstack command hung")
+        else:
+            if err:
+                self.fail("STDERR: " + err.decode())
+
 
 if __name__ == "__main__":
     unittest.main()
