@@ -1,6 +1,6 @@
 #!/usr/bin/env python
 #
-# Copyright (c) 2024, Ryan Galloway (ryan@rsgalloway.com)
+# Copyright (c) 2024-2025, Ryan Galloway (ryan@rsgalloway.com)
 #
 # Redistribution and use in source and binary forms, with or without
 # modification, are permitted provided that the following conditions are met:
@@ -57,7 +57,7 @@ def create_test_root():
     for env in ("prod", "dev"):
         shutil.copytree(env_path, os.path.join(root, env, "env"))
 
-    return root
+    return root.replace("\\", "/")  # normalize path for windows
 
 
 def update_env_file(file_path: str, key: str, value: str):
@@ -172,6 +172,7 @@ class TestEnv(unittest.TestCase):
     def test_bake(self):
         """Tests baking an environment."""
         from envstack.env import load_environ
+
         env = load_environ("thing")
         baked = env.bake()
         for k, v in env.items():
@@ -182,6 +183,7 @@ class TestEnv(unittest.TestCase):
     def test_bake_out(self):
         """Tests bake, write and load an environment."""
         from envstack.env import load_environ
+
         env1 = load_environ("thing")
         self.filename = "test_bake_out.env"
         env1.write(self.filename)
@@ -194,6 +196,7 @@ class TestEnv(unittest.TestCase):
     def test_write_simple(self):
         """Tests writing an environment to a file."""
         from envstack.env import load_environ, resolve_environ
+
         env1 = Env({"FOO": "foo", "BAR": "${FOO}"})
         self.filename = "test_write_simple.env"
         env1.write(self.filename)
@@ -210,6 +213,7 @@ class TestEnv(unittest.TestCase):
         """Tests writing an environment with a custom node to a file."""
         from envstack.env import load_environ, resolve_environ
         from envstack.node import EncryptedNode
+
         env1 = Env({"FOO": "foo", "BAR": EncryptedNode("bar")})
         self.filename = "test_write_custom.env"
         # write it out and reload it
@@ -222,8 +226,8 @@ class TestEnv(unittest.TestCase):
         self.assertEqual(env1["FOO"], "foo")
         self.assertEqual(env1["BAR"], EncryptedNode("bar"))
         self.assertEqual(env2["FOO"], "foo")
-        self.assertEqual(env2["BAR"], EncryptedNode('YmFy'))
-        self.assertEqual(EncryptedNode('YmFy').resolve(env=env2), "bar")
+        self.assertEqual(env2["BAR"], EncryptedNode("YmFy"))
+        self.assertEqual(EncryptedNode("YmFy").resolve(env=env2), "bar")
         self.assertEqual(env3["FOO"], "foo")
         self.assertEqual(env3["BAR"], "bar")
         self.assertEqual(env4["FOO"], "foo")
@@ -364,6 +368,26 @@ class TestInit(unittest.TestCase):
 class TestResolveEnviron(unittest.TestCase):
     """Tests the resolve_environ function."""
 
+    def test_home(self):
+        """Tests to make sure ${HOME} is resolved."""
+        from envstack.env import resolve_environ
+
+        # ${HOME} is undefined on windows
+        home = os.getenv("HOME", os.path.expanduser("~"))
+        os.environ["HOME"] = home
+        env = {"FOO": "${HOME}/foo"}
+        resolved = resolve_environ(env)
+        self.assertEqual(resolved["FOO"], f"{home}/foo")
+
+    def test_custom(self):
+        """Tests to make sure ${CUSTOM} is resolved from os.environ."""
+        from envstack.env import resolve_environ
+
+        os.environ["CUSTOM"] = "/var/tmp"
+        env = {"FOO": "${CUSTOM}/foo"}
+        resolved = resolve_environ(env)
+        self.assertEqual(resolved["FOO"], f"/var/tmp/foo")
+
     def test_simple(self):
         """Tests resolving a simple environment."""
         from envstack.env import resolve_environ
@@ -457,7 +481,7 @@ class TestResolveEnviron(unittest.TestCase):
         os.environ["ROOT"] = "/mnt/pipe"
         os.environ["ENV"] = "dev"
         env = {
-            "DEPLOY_ROOT": "${ROOT}/${ENV}}",
+            "DEPLOY_ROOT": "${ROOT}/${ENV}",
             "ENV": "${ENV:=prod}",
             "ROOT": "${ROOT:=/var/tmp}",
         }
@@ -470,7 +494,7 @@ class TestResolveEnviron(unittest.TestCase):
 
         env_value = os.getenv("ENV", "prod")
         env = {
-            "DEPLOY_ROOT": "${MOUNT}/${DRIVE}/${ENV}}",
+            "DEPLOY_ROOT": "${MOUNT}/${DRIVE}/${ENV}",
             "ENV": "${ENV:=prod}",
             "MOUNT": "/mnt",
             "DRIVE": "${DRIVE:=pipe}",
@@ -539,30 +563,40 @@ class TestResolveEnviron(unittest.TestCase):
         """Tests expansion modifier with deferred value."""
         from envstack.env import resolve_environ
 
-        env = {"VAR": "${VAR:=${FOO}}", "FOO": "${FOO:=/foo/bar}"}
-        env["BAR"] = "${BAZ}"
+        env = {
+            "VAR": "${VAR:=${FOO}}",
+            "FOO": "${FOO:=/foo/bar}",
+            "BAR": "${BAZ}",  # has emptry string value
+        }
         resolved = resolve_environ(env)
         self.assertEqual(resolved["VAR"], "/foo/bar")
         self.assertEqual(resolved["FOO"], "/foo/bar")
+        self.assertEqual(resolved["BAR"], "")
 
     def test_expansion_modifier_deferred_default(self):
         """Tests expansion modifier with multiple deferred values."""
         from envstack.env import resolve_environ
 
-        env = {"VAR": "${VAR:=${FOO}}", "FOO": "${FOO:=${BAR}}"}
-        env["BAR"] = "${BAZ:=/baz/qux}"  # insert a default value
+        env = {
+            "VAR": "${VAR:=${FOO}}",
+            "FOO": "${FOO:=${BAR}}",
+            "BAR": "${BAZ:=/baz/qux}",  # has a default value
+        }
         resolved = resolve_environ(env)
         self.assertEqual(resolved["VAR"], "/baz/qux")
         self.assertEqual(resolved["FOO"], "/baz/qux")
         self.assertEqual(resolved["BAR"], "/baz/qux")
 
-    def test_expansion_modifier_deferred_default_slash(self):
-        """Tests expansion modifier with multiple deferred values."""
+    def test_expansion_modifier_deferred_null_value(self):
+        """Tests expansion modifier with multiple deferred values and empty string."""
         from envstack.env import resolve_environ
 
-        env = {"VAR": "${VAR:=${FOO}}", "FOO": "${FOO:=${BAR}}"}
-        env["BAR"] = "${BAZ:=/baz/qux}"
-        env["BAZ"] = "${QUX}"
+        env = {
+            "VAR": "${VAR:=${FOO}}",
+            "FOO": "${FOO:=${BAR}}",
+            "BAR": "${BAZ:=/baz/qux}",
+            "BAZ": "${QUX}",  # has empty string value
+        }
         resolved = resolve_environ(env)
         self.assertEqual(resolved["VAR"], "/baz/qux")
         self.assertEqual(resolved["FOO"], "/baz/qux")
@@ -966,6 +1000,204 @@ class TestIssues(unittest.TestCase):
         self.assertTrue("yaml" in sys.modules)
         self.assertTrue(len(sys.path) > 0)
 
+    def test_issue_51(self):
+        """Tests issue #51 with URL values on all platforms."""
+        from envstack.env import load_environ
+
+        env = Env(
+            {
+                "URL": "https://example.com",
+                "S3": "s3://bucket.amazonaws.com",
+                "GIT": "git://path/to/repo.git",
+            }
+        )
+        self.filename = os.path.join(self.root, "test_issue_51.env")
+        env.write(self.filename)
+        self.assertEqual(env["URL"], "https://example.com")
+        self.assertEqual(env["S3"], "s3://bucket.amazonaws.com")
+        self.assertEqual(env["GIT"], "git://path/to/repo.git")
+
+        for platform in ("linux", "windows", "darwin"):
+            env = load_environ(self.filename, platform=platform)
+            self.assertEqual(env["URL"], "https://example.com")
+            self.assertEqual(env["S3"], "s3://bucket.amazonaws.com")
+            self.assertEqual(env["GIT"], "git://path/to/repo.git")
+
+    def test_issue_55(self):
+        """Tests issue #55 with missing environment file.
+        Tests dynamic ${ENVPATH} values that use ${STACK} in the path,
+
+        env/test.env:
+            ENVPATH: ${ROOT}/${STACK}/env:${ROOT}/prod/env
+
+        The STACK name and the test env file name should be the same.
+        """
+        from envstack.env import load_environ, Env
+
+        # update default.env to point to test root
+        default_env_file = os.path.join(self.root, "prod", "env", "default.env")
+        dev_env_file = os.path.join(self.root, "prod", "env", "dev.env")
+        update_env_file(default_env_file, "ROOT", self.root)
+
+        # create a new test env file that only exists in our test env dir
+        test_env = os.path.join(self.root, "test_issue_55", "env")
+        os.makedirs(test_env, exist_ok=True)
+        test_env_file = os.path.join(test_env, "test_issue_55.env")
+
+        data = {"FOO": "foo", "BAR": "bar", "BAZ": self.root}
+        Env(data).write(test_env_file)
+        self.assertTrue(os.path.exists(test_env_file))
+
+        # try to load our test env file by loading the "dev" env first, which
+        # does not use STACK in ENVPATH
+        env1 = load_environ(["dev", "test_issue_55"])
+
+        # last env file should be the "dev" env file
+        self.assertEqual(str(env1.sources[-1].path), dev_env_file)
+        self.assertRaises(KeyError, lambda: env1["FOO"])
+        self.assertRaises(KeyError, lambda: env1["BAR"])
+        self.assertRaises(KeyError, lambda: env1["BAZ"])
+
+        # FIXME: why is this necessary? (think it's caching seen stacks)
+        envstack.revert()
+
+        # load our test env file by loading the "test" env first, which
+        # does use STACK in ENVPATH
+        env2 = load_environ(["test", "test_issue_55"])
+
+        # last env file should be our test env file
+        self.assertEqual(str(env2.sources[-1].path), test_env_file)
+        self.assertEqual(env2["FOO"], "foo")
+        self.assertEqual(env2["BAR"], "bar")
+        self.assertEqual(env2["BAZ"], self.root)
+        self.assertEqual(env2["STACK"], "test_issue_55")
+
+    def test_issue_58(self):
+        """Tests for issue #58 and #62, inherited environment variables.
+        Tests different kinds of inheritance, values and expansion modifiers.
+
+        grandparent:
+            FOO: grandparent
+            BAR: ${BAR:=c}
+            BAZ: baz
+            NUM: 2
+        parent:
+            include: [grandparent]
+            FOO: ${FOO:=parent}
+            BAR: ${BAR:=b}
+            BAZ: ${BAZ}
+            NUM: ${NUM:=1}
+        child:
+            include: [parent]
+            FOO: ${FOO:=child}
+            BAR: ${BAR:=a}
+            BAZ: ${BAZ}
+            NUM: ${NUM:=0}
+        """
+        from envstack.env import load_environ, resolve_environ, Source
+
+        # create grandparent.env that sets a value for FOO
+        grandparent = {
+            "include": [],
+            "all": {"FOO": "grandparent", "BAR": "${BAR:=c}", "BAZ": "baz", "NUM": "2"},
+        }
+        grandparent_env_file = os.path.join(self.root, "prod", "env", "grandparent.env")
+        grandparent_source = Source(grandparent_env_file)
+        grandparent_source.data = grandparent
+        grandparent_source.write()
+
+        # create parent.env that includes grandparent
+        parent = {
+            "include": ["grandparent"],
+            "all": {
+                "FOO": "${FOO:=parent}",
+                "BAR": "${BAR:=b}",
+                "BAZ": "${BAZ}",
+                "INT": "1",
+                "NUM": "${NUM:=1}",
+            },
+        }
+        parent_env_file = os.path.join(self.root, "prod", "env", "parent.env")
+        parent_source = Source(parent_env_file)
+        parent_source.data = parent
+        parent_source.write()
+        parent_env = load_environ("parent")
+        self.assertEqual(parent_env["INT"], 1)
+
+        # FIXME: need to revert to clear the cached environment
+        envstack.revert()
+
+        # create child.env that includes parent
+        child = {
+            "include": ["parent"],
+            "all": {
+                "FOO": "${FOO:=child}",
+                "BAR": "${BAR:=a}",
+                "BAZ": "${BAZ}",
+                "INT": "${INT}",
+                "NUM": "${NUM:=0}",
+                "TEST": "foo",
+            },
+        }
+        child_env_file = os.path.join(self.root, "prod", "env", "child.env")
+        child_source = Source(child_env_file)
+        child_source.data = child
+        child_source.write()
+
+        env = load_environ("child")
+        resolved = resolve_environ(env)
+        self.assertEqual(len(env.sources), 3)
+        self.assertEqual(env["FOO"], "${FOO:=child}")
+        self.assertEqual(env["BAR"], "${BAR:=a}")
+        self.assertEqual(env["BAZ"], "${BAZ}")
+        self.assertEqual(env["NUM"], "${NUM:=0}")
+        self.assertEqual(env["INT"], "${INT}")
+        self.assertEqual(env["TEST"], "foo")
+        self.assertEqual(resolved["FOO"], "grandparent")
+        self.assertEqual(resolved["BAR"], "c")
+        self.assertEqual(resolved["BAZ"], "baz")
+        self.assertEqual(resolved["NUM"], 2)
+        self.assertEqual(resolved["INT"], 1)
+        self.assertEqual(resolved["TEST"], "foo")
+
+        # simulate a new process, and init the environment
+        envstack.revert()
+        envstack.init("child")
+
+        # os.environ expects str values
+        self.assertEqual(os.getenv("FOO"), "grandparent")
+        self.assertEqual(os.getenv("BAR"), "c")
+        self.assertEqual(os.getenv("BAZ"), "baz")
+        self.assertEqual(os.getenv("INT"), "1")
+        self.assertEqual(os.getenv("NUM"), "2")
+        self.assertEqual(os.getenv("TEST"), "foo")
+
+    def test_issue_59(self):
+        """Tests for issue #59: multiple levels of indirect references."""
+        from envstack.env import resolve_environ
+
+        # two levels of indirect references with defaults
+        environ = {"VAR": "${FOO:=var}", "FOO": "${BAR:=foo}", "BAR": "bar"}
+        resolved = resolve_environ(environ)
+        self.assertEqual(resolved["VAR"], "bar")
+        self.assertEqual(resolved["FOO"], "bar")
+        self.assertEqual(resolved["BAR"], "bar")
+
+        # four levels of indirect references w/o defaults
+        environ = {
+            "VAR": "${FOO}",
+            "FOO": "${BAR}",
+            "BAR": "${BAZ}",
+            "BAZ": "${QUX}",
+            "QUX": "foo",
+        }
+        resolved = resolve_environ(environ)
+        self.assertEqual(resolved["VAR"], "foo")
+        self.assertEqual(resolved["FOO"], "foo")
+        self.assertEqual(resolved["BAR"], "foo")
+        self.assertEqual(resolved["BAZ"], "foo")
+        self.assertEqual(resolved["QUX"], "foo")
+
 
 if __name__ == "__main__":
     unittest.main()
